#! /usr/bin/env python3

from .base_metric_loss_function import BaseMetricLossFunction
import torch
from ..utils import loss_and_miner_utils as lmu, common_functions as c_f


class MarginLoss(BaseMetricLossFunction):

    def __init__(self, margin, nu, beta, **kwargs):
        self.margin = margin
        self.nu = nu
        self.beta = beta
        self.num_pos_pairs = 0
        self.num_neg_pairs = 0
        self.record_these = ["num_pos_pairs", "num_neg_pairs"]
        super().__init__(**kwargs)

    def compute_loss(self, embeddings, labels, indices_tuple):
        anchor_idx, positive_idx, negative_idx = lmu.convert_to_triplets(indices_tuple, labels)
        if len(anchor_idx) == 0:
            self.num_pos_pairs = 0
            self.num_neg_pairs = 0
            return 0
        anchors, positives, negatives = embeddings[anchor_idx], embeddings[positive_idx], embeddings[negative_idx]
        beta = self.maybe_mask_param(self.beta, labels[anchor_idx])
        beta_reg_loss = self.compute_reg_loss(beta)

        d_ap = torch.nn.functional.pairwise_distance(positives, anchors, p=2)
        d_an = torch.nn.functional.pairwise_distance(negatives, anchors, p=2)

        pos_loss = torch.nn.functional.relu(d_ap - beta + self.margin)
        neg_loss = torch.nn.functional.relu(beta - d_an + self.margin)

        self.num_pos_pairs = (pos_loss > 0.0).nonzero().size(0)
        self.num_neg_pairs = (neg_loss > 0.0).nonzero().size(0)

        pair_count = self.num_pos_pairs + self.num_neg_pairs 

        return (torch.sum(pos_loss + neg_loss) + beta_reg_loss) / (pair_count + 1e-16)

    def compute_reg_loss(self, beta):
        if self.nu > 0:
            beta_mean = c_f.try_torch_operation(torch.mean, beta)
            return beta_mean * self.nu
        return 0


import collections
import torch
from torch.autograd import Variable
import numpy as np



def try_keys(input_dict, keys):
    for k in keys:
        try:
            return input_dict[k]
        except BaseException:
            pass
    return None


def try_next_on_generator(gen, iterable):
    try:
        return gen, next(gen)
    except StopIteration:
        gen = iter(iterable)
        return gen, next(gen)


def apply_func_to_dict(input, f):
    if isinstance(input, collections.Mapping):
        for k, v in input.items():
            input[k] = f(v)
        return input
    else:
        return f(input)


def wrap_variable(batch_data, device):
    def f(x):
        return Variable(x).to(device)

    return apply_func_to_dict(batch_data, f)


def get_hierarchy_label(batch_labels, hierarchy_level):
    def f(v):
        try:
            if v.ndim == 2:
                v = v[:, hierarchy_level]
            return v
        except BaseException:
            return v

    return apply_func_to_dict(batch_labels, f)


def numpy_to_torch(input):
    def f(v):
        try:
            return torch.from_numpy(v)
        except BaseException:
            return v

    return apply_func_to_dict(input, f)


def torch_to_numpy(input):
    def f(v):
        try:
            return v.cpu().numpy()
        except BaseException:
            return v

    return apply_func_to_dict(input, f)


def process_label(labels, hierarchy_level, label_map):
    labels = get_hierarchy_label(labels, hierarchy_level)
    labels = torch_to_numpy(labels)
    labels = label_map(labels, hierarchy_level)
    labels = numpy_to_torch(labels)
    return labels


def pass_data_to_model(model, data, device, **kwargs):
    if isinstance(data, collections.Mapping):
        base_output = {}
        for k, v in data.items():
            base_output[k] = model(wrap_variable(v, device), k=k, **kwargs)
        return base_output
    else:
        return model(wrap_variable(data, device), **kwargs)

def set_requires_grad(model, requires_grad):
    for param in model.parameters():
        param.requires_grad = requires_grad



def copy_params_to_another_model(from_model, to_model):
    params1 = from_model.named_parameters()
    params2 = to_model.named_parameters()
    dict_params2 = dict(params2)
    for name1, param1 in params1:
        if name1 in dict_params2:
            dict_params2[name1].data.copy_(param1.data)


def safe_random_choice(input_data, size):
    """
    Randomly samples without replacement from a sequence. It is "safe" because
    if len(input_data) < size, it will randomly sample WITH replacement
    Args:
        input_data is a sequence, like a torch tensor, numpy array,
                        python list, tuple etc
        size is the number of elements to randomly sample from input_data
    Returns:
        An array of size "size", randomly sampled from input_data
    """
    replace = len(input_data) < size
    return np.random.RandomState().choice(input_data, size=size, replace=replace)


def longest_list(list_of_lists):
    return max(list_of_lists, key=len)


def slice_by_n(input_array, n):
    output = []
    for i in range(n):
        output.append(input_array[i::n])
    return output


def unslice_by_n(input_tensors):
    n = len(input_tensors)
    rows, cols = input_tensors[0].size()
    output = torch.zeros((rows * n, cols)).to(input_tensors[0].device)
    for i in range(n):
        output[i::n] = input_tensors[i]
    return output


def set_layers_to_eval(layer_name):
    def set_to_eval(m):
        classname = m.__class__.__name__
        if classname.find(layer_name) != -1:
            m.eval()
    return set_to_eval


def get_dataloader(dataset, batch_size, sampler, num_workers, collate_fn):
    return torch.utils.data.DataLoader(
        dataset,
        batch_size=int(batch_size),
        sampler=sampler,
        drop_last=True,
        num_workers=num_workers,
        collate_fn=collate_fn,
        shuffle=sampler is None
    )


def try_torch_operation(torch_op, input_val):
    return torch_op(input_val) if torch.is_tensor(input_val) else input_val 


#! /usr/bin/env python3

from .metric_loss_only import MetricLossOnly


class TrainWithClassifier(MetricLossOnly):
    def loss_names(self):
        return ["metric_loss", "classifier_loss"]

    def calculate_loss(self, curr_batch):
        data, labels = curr_batch
        embeddings, labels = self.compute_embeddings(data, labels)
        logits = self.maybe_get_logits(embeddings)
        indices_tuple = self.maybe_mine_embeddings(embeddings, labels)
        self.losses["metric_loss"] = self.maybe_get_metric_loss(embeddings, labels, indices_tuple)
        self.losses["classifier_loss"] = self.maybe_get_classifier_loss(logits, labels)

    def maybe_get_classifier_loss(self, logits, labels):
        if logits is not None:
            return self.loss_funcs["classifier_loss"](logits, labels.to(logits.device))
        return 0

    def maybe_get_logits(self, embeddings):
        if self.models.get("classifier", None) and self.loss_weights.get("classifier_loss", 0) > 0:
            return self.models["classifier"](embeddings)
        return None

        
        
        
        



#! /usr/bin/env python3

from .base_metric_loss_function import BaseMetricLossFunction
import numpy as np
import torch
from ..utils import loss_and_miner_utils as lmu


class AngularNPairsLoss(BaseMetricLossFunction):
    """
    Implementation of https://arxiv.org/abs/1708.01682
    Args:
        alpha: The angle (as described in the paper), specified in degrees.
    """
    def __init__(self, alpha, **kwargs):
        self.alpha = torch.tensor(np.radians(alpha))
        self.maybe_modify_loss = lambda x: x
        self.num_anchors = 0
        self.avg_embedding_norm = 0
        self.record_these = ["num_anchors", "avg_embedding_norm"]
        super().__init__(**kwargs)

    def compute_loss(self, embeddings, labels, indices_tuple):
        self.avg_embedding_norm = torch.mean(torch.norm(embeddings, p=2, dim=1))
        anchor_idx, positive_idx = lmu.convert_to_pos_pairs_with_unique_labels(indices_tuple, labels)
        self.num_anchors = len(anchor_idx)
        if self.num_anchors == 0:
            return 0

        anchors, positives = embeddings[anchor_idx], embeddings[positive_idx]
        points_per_anchor = (self.num_anchors - 1) * 2

        alpha = self.maybe_mask_param(self.alpha, labels[anchor_idx])
        sq_tan_alpha = torch.tan(alpha) ** 2

        xa_xp = torch.sum(anchors * positives, dim=1, keepdim=True)
        term2_multiplier = 2 * (1 + sq_tan_alpha)
        term2 = term2_multiplier * xa_xp

        a_p_summed = anchors + positives
        inside_exp = []
        mask = torch.ones(self.num_anchors).to(embeddings.device) - torch.eye(self.num_anchors).to(embeddings.device)
        term1_multiplier = 4 * sq_tan_alpha

        for negatives in [anchors, positives]:
            term1 = term1_multiplier * torch.matmul(a_p_summed, torch.t(negatives))
            inside_exp.append(term1 - term2.repeat(1, self.num_anchors))
            inside_exp[-1] = inside_exp[-1] * mask

        inside_exp_final = torch.zeros((self.num_anchors, points_per_anchor + 1)).to(embeddings.device)

        for i in range(self.num_anchors):
            indices = np.concatenate((np.arange(0, i), np.arange(i + 1, inside_exp[0].size(1))))
            inside_exp_final[i, : points_per_anchor // 2] = inside_exp[0][i, indices]
        inside_exp_final[:, points_per_anchor // 2 :] = inside_exp[1]
        inside_exp_final = self.maybe_modify_loss(inside_exp_final)

        return torch.mean(torch.logsumexp(inside_exp_final, dim=1))

    def create_learnable_parameter(self, init_value):
        return super().create_learnable_parameter(init_value, unsqueeze=True)


class AngularLoss(AngularNPairsLoss):
    def compute_loss(self, embeddings, labels, indices_tuple):
        self.avg_embedding_norm = torch.mean(torch.norm(embeddings, p=2, dim=1))
        anchor_idx, positive_idx, negative_idx = lmu.convert_to_triplets(indices_tuple, labels)
        self.num_anchors = len(anchor_idx)
        if self.num_anchors == 0:
            return 0
        anchors, positives, negatives = embeddings[anchor_idx], embeddings[positive_idx], embeddings[negative_idx]
        alpha = self.maybe_mask_param(self.alpha, labels[anchor_idx])
        sq_tan_alpha = torch.tan(alpha) ** 2
        term1 = 4 * sq_tan_alpha * torch.sum((anchors + positives) * negatives, dim=1, keepdim=True)
        term2 = 2 * (1 + sq_tan_alpha) * torch.sum(anchors * positives, dim=1, keepdim=True)
        final_form = torch.cat([term1 - term2, torch.zeros(term1.size(0), 1).to(embeddings.device)], dim=1)
        final_form = self.maybe_modify_loss(final_form)
        return torch.mean(torch.logsumexp(final_form, dim=1))



#! /usr/bin/env python3

from .base_metric_loss_function import BaseMetricLossFunction
from ..utils import loss_and_miner_utils as lmu
import torch


class NCALoss(BaseMetricLossFunction):
    # modified from https://github.com/microsoft/snca.pytorch/blob/master/lib/NCA.py
    # https://www.cs.toronto.edu/~hinton/absps/nca.pdf
    def compute_loss(self, embeddings, labels, *_):
        return self.nca_computation(embeddings, embeddings, labels, labels)

    def nca_computation(self, query, reference, query_labels, reference_labels):
        query_batch_size = len(query)
        reference_batch_size = len(reference)
        x = lmu.dist_mat(query, reference, squared=True)
        exp = torch.exp(-x)
        
        if query is reference:
            exp = exp - torch.diag(exp.diag())
        repeated_labels = query_labels.view(query_batch_size, 1).repeat(1, reference_batch_size)
        same_labels = (repeated_labels == reference_labels).float()
        
        # sum over all positive neighbors of each anchor
        p = torch.sum(exp * same_labels, dim=1)
        # sum over all neighbors of each anchor (excluding the anchor)
        Z = torch.sum(exp * (1 - same_labels), dim=1)
        prob = p / Z
        non_zero_prob = torch.masked_select(prob, prob != 0)
        
        return -torch.mean(torch.log(non_zero_prob))


#! /usr/bin/env python3

from .nca_loss import NCALoss
import torch
from ..utils import loss_and_miner_utils as lmu

class ProxyNCALoss(NCALoss):
    def __init__(self, num_labels, embedding_size, **kwargs):
        self.proxies = torch.nn.Parameter(torch.randn(num_labels, embedding_size))
        self.proxy_labels = torch.arange(num_labels)
        super().__init__(**kwargs)

    def compute_loss(self, embeddings, labels, *_):
        if self.normalize_embeddings:
            prox = torch.nn.functional.normalize(self.proxies, p=2, dim=1)
        else:
            prox = self.proxies
        return self.nca_computation(embeddings, prox, labels, self.proxy_labels.to(labels.device))


#! /usr/bin/env python3
from .base_miner import BasePostGradientMiner
import torch
from ..utils import loss_and_miner_utils as lmu


# mining method used in Hard Aware Deeply Cascaded Embeddings
# https://arxiv.org/abs/1611.05720
class HDCMiner(BasePostGradientMiner):
    def __init__(self, filter_amounts, use_sim_mat=False, **kwargs):
        super().__init__(**kwargs)
        self.num_pos_pairs_per_round = torch.zeros(len(filter_amounts))
        self.num_neg_pairs_per_round = torch.zeros(len(filter_amounts))
        self.record_these = ["num_pos_pairs_per_round", "num_neg_pairs_per_round"]
        self.filter_amounts = filter_amounts
        self.use_sim_mat = use_sim_mat
        self.i = 0
        self.reset_prev_idx()

    def mine(self, embeddings, labels):
        if self.i == 0:
            self.num_pos_pairs_per_round *= 0
            self.num_neg_pairs_per_round *= 0
            if self.use_sim_mat:
                self.sim_mat = lmu.sim_mat(embeddings)
            else:
                self.dist_mat = lmu.dist_mat(embeddings, squared=False)
            self.a1_idx, self.p_idx, self.a2_idx, self.n_idx = lmu.get_all_pairs_indices(
                labels
            )
            self.reset_prev_idx()

        self.maybe_set_to_prev()
        curr_filter = self.filter_amounts[self.i]
        if curr_filter != 1:
            mat = self.sim_mat if self.use_sim_mat else self.dist_mat
            pos_pair_ = mat[self.a1_idx, self.p_idx]
            neg_pair_ = mat[self.a2_idx, self.n_idx]

            a1, p, a2, n = [], [], [], []

            for name, v in {"pos": pos_pair_, "neg": neg_pair_}.items():
                num_pairs = len(v)
                k = int(curr_filter * num_pairs)
                largest = self.should_select_largest(name)
                _, idx = torch.topk(v, k=k, largest=largest)
                self.append_original_indices(name, idx, a1, p, a2, n)

            self.a1_idx = torch.cat(a1)
            self.p_idx = torch.cat(p)
            self.a2_idx = torch.cat(a2)
            self.n_idx = torch.cat(n)

        self.num_pos_pairs_per_round[self.i] = len(self.a1_idx)
        self.num_neg_pairs_per_round[self.i] = len(self.a2_idx)
        self.set_prev_idx()
        self.i = (self.i + 1) % len(self.filter_amounts)
        return self.a1_idx, self.p_idx, self.a2_idx, self.n_idx

    def should_select_largest(self, name):
        if self.use_sim_mat:
            return False if name == "pos" else True
        return True if name == "pos" else False

    def append_original_indices(self, name, idx, a1, p, a2, n):
        if name == "pos":
            a1.append(self.a1_idx[idx])
            p.append(self.p_idx[idx])
        else:
            a2.append(self.a2_idx[idx])
            n.append(self.n_idx[idx])

    def maybe_set_to_prev(self):
        if self.prev_a1 is not None:
            self.a1_idx = self.prev_a1
            self.p_idx = self.prev_p
            self.a2_idx = self.prev_a2
            self.n_idx = self.prev_n

    def reset_prev_idx(self):
        self.prev_a1 = None
        self.prev_p = None
        self.prev_a2 = None
        self.prev_n = None

    def set_prev_idx(self, reset=False):
        self.prev_a1 = self.a1_idx.clone()
        self.prev_p = self.p_idx.clone()
        self.prev_a2 = self.a2_idx.clone()
        self.prev_n = self.n_idx.clone()



#! /usr/bin/env python3

import torch


class BaseMiner(torch.nn.Module):
    def __init__(self, normalize_embeddings=True):
        super().__init__()
        self.normalize_embeddings = normalize_embeddings

    def mine(self, embeddings, labels):
        """
        Args:
            embeddings: tensor of size (batch_size, embedding_size)
            labels: tensor of size (batch_size)
        Returns: a tuple where each element is an array of indices.
        """
        raise NotImplementedError

    def output_assertion(self, output):
        raise NotImplementedError

    def forward(self, embeddings, labels):
        """
        Args:
            embeddings: tensor of size (batch_size, embedding_size)
            labels: tensor of size (batch_size)
        Does any necessary preprocessing, then does mining, and then checks the
        shape of the mining output before returning it
        """
        labels = labels.to(embeddings.device)
        with torch.no_grad():
            if self.normalize_embeddings:
                embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=1) 
            mining_output = self.mine(embeddings, labels)
        self.output_assertion(mining_output)
        return mining_output

class BasePostGradientMiner(BaseMiner):
    """
    A post-gradient miner is used after gradients have already been computed. 
    In other words, the composition of the batch has already been decided, 
    and the miner will find pairs or triplets within the batch that should 
    be used to compute the loss.
    Args:
        normalize_embeddings: type boolean, if True then normalize embeddings
                                to have norm = 1 before mining
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.num_pos_pairs = 0
        self.num_neg_pairs = 0
        self.num_triplets = 0
        record_these = ["num_pos_pairs", "num_neg_pairs", "num_triplets"]
        if hasattr(self, "record_these"):
            self.record_these += record_these
        else:
            self.record_these = record_these


    def output_assertion(self, output):
        """
        Args:
            output: the output of self.mine
        This asserts that the mining function is outputting
        properly formatted indices. The default is to require a tuple representing
        a,p,n indices or a1,p,a2,n indices within a batch of embeddings.
        For example, a tuple of (anchors, positives, negatives) will be
        (torch.tensor, torch.tensor, torch.tensor)
        """
        if len(output) == 3:
            self.num_triplets = len(output[0])
            assert self.num_triplets == len(output[1]) == len(output[2])
        elif len(output) == 4:
            self.num_pos_pairs = len(output[0])
            self.num_neg_pairs = len(output[2])
            assert self.num_pos_pairs == len(output[1])
            assert self.num_neg_pairs == len(output[3])
        else:
            raise BaseException


class BasePreGradientMiner(BaseMiner):
    """
    A pre-gradient miner is used before gradients have been computed.
    The miner finds a subset of the sampled batch for which gradients will
    then need to be computed.
    Args:
        output_batch_size: type int. The size of the subset that the miner
                            will output.
        normalize_embeddings: type boolean, if True then normalize embeddings
                                to have norm = 1 before mining
    """

    def __init__(self, output_batch_size, **kwargs):
        super().__init__(**kwargs)
        self.output_batch_size = output_batch_size

    def output_assertion(self, output):
        assert len(output) == self.output_batch_size


#! /usr/bin/env python3
import faiss
import torch
import logging

# modified from https://github.com/facebookresearch/deepcluster
def get_knn(
    reference_embeddings, test_embeddings, k, embeddings_come_from_same_source=False
):
    """
    Finds the k elements in reference_embeddings that are closest to each
    element of test_embeddings.
    Args:
        reference_embeddings: numpy array of size (num_samples, dimensionality).
        test_embeddings: numpy array of size (num_samples2, dimensionality).
        k: int, number of nearest neighbors to find
        embeddings_come_from_same_source: if True, then the nearest neighbor of
                                         each element (which is actually itself)
                                         will be ignored.
    """
    logging.info("running k-nn with k=%d"%k)
    d = reference_embeddings.shape[1]
    res = faiss.StandardGpuResources()
    flat_config = faiss.GpuIndexFlatConfig()
    flat_config.device = int(torch.cuda.device_count()) - 1
    index = faiss.GpuIndexFlatL2(res, d, flat_config)
    index.add(reference_embeddings)
    _, indices = index.search(test_embeddings, k + 1)
    if embeddings_come_from_same_source:
        return indices[:, 1:]
    return indices[:, :k]


# from https://raw.githubusercontent.com/facebookresearch/deepcluster/
def run_kmeans(x, nmb_clusters):
    """Runs kmeans on 1 GPU.
    Args:
        x: data
        nmb_clusters (int): number of clusters
    Returns:
        list: ids of data in each cluster
    """
    n_data, d = x.shape

    # faiss implementation of k-means
    clus = faiss.Clustering(d, nmb_clusters)
    clus.niter = 20
    clus.max_points_per_centroid = 10000000
    res = faiss.StandardGpuResources()
    flat_config = faiss.GpuIndexFlatConfig()
    flat_config.useFloat16 = False
    flat_config.device = 0
    index = faiss.GpuIndexFlatL2(res, d, flat_config)

    # perform the training
    clus.train(x, index)
    _, idxs = index.search(x, 1)

    return [int(n[0]) for n in idxs]



#! /usr/bin/env python3

import tqdm

import torch
import numpy as np
from ..utils import stat_utils
from ..utils import common_functions as c_f
import logging



class BaseTester:
    def __init__(self, reference_set="compared_to_self", normalize_embeddings=True, use_trunk_output=False, 
                    batch_size=32, dataloader_num_workers=32, metric_for_best_epoch="ordered_r_precision", 
                    data_device=None, record_keeper=None):
        self.reference_set = reference_set
        self.normalize_embeddings = normalize_embeddings
        self.use_trunk_output = use_trunk_output
        self.batch_size = int(batch_size)
        self.key_for_best_epoch = self.accuracies_keyname(metric_for_best_epoch, 0)
        self.data_device = torch.device("cuda" if torch.cuda.is_available() else "cpu") if data_device is None else data_device
        self.num_workers = dataloader_num_workers
        self.record_keeper = record_keeper
        self.base_record_group_name = self.suffixes("%s_%s"%("accuracies", self.__class__.__name__))

    def get_best_epoch_and_accuracy(self, split_name):
        records = self.record_keeper.get_record(self.record_group_name(split_name))
        accuracies = records[self.key_for_best_epoch]
        return records["epoch"][np.argmax(accuracies)], np.max(accuracies)

    def maybe_normalize(self, embeddings):
        if self.normalize_embeddings:
            return embeddings / np.expand_dims(np.linalg.norm(embeddings, axis=1), 1)
        return embeddings

    def compute_all_embeddings(self, dataloader, trunk_model, embedder_model, post_processor):
        num_batches = len(dataloader)
        s, e = 0, 0
        with torch.no_grad():
            for i, data in enumerate(tqdm.tqdm(dataloader)):
                input_imgs = c_f.try_keys(data, ["data", "image"])
                label = data["label"]
                q = self.get_embeddings_for_eval(trunk_model, embedder_model, input_imgs)
                label = c_f.numpy_to_torch(label)
                q, label = post_processor(q, label)
                if label.dim() == 1:
                    label = label.unsqueeze(1)
                if i == 0:
                    labels = torch.zeros(len(dataloader.dataset), label.size(1))
                    all_q = torch.zeros(len(dataloader.dataset), q.size(1))
                e = s + q.size(0)
                all_q[s:e] = q
                labels[s:e] = label
                s = e
            labels = labels.cpu().numpy()
            all_q = all_q.cpu().numpy()

        return all_q, labels

    def get_all_embeddings(self, dataset, trunk_model, embedder_model, post_processor, collate_fn):
        dataloader = torch.utils.data.DataLoader(
            dataset,
            batch_size=self.batch_size,
            shuffle=False,
            drop_last=False,
            collate_fn=collate_fn,
            num_workers=self.num_workers,
        )
        if post_processor is None: post_processor = lambda e, l: (e, l)
        embeddings, labels = self.compute_all_embeddings(dataloader, trunk_model, embedder_model, post_processor)
        embeddings = self.maybe_normalize(embeddings)
        return embeddings, labels

    def get_embeddings_for_eval(self, trunk_model, embedder_model, input_imgs):
        trunk_output = c_f.pass_data_to_model(trunk_model, input_imgs, self.data_device)
        if self.use_trunk_output:
            return trunk_output
        return embedder_model(trunk_output)

    def suffixes(self, base_name):
        if self.normalize_embeddings:
            base_name += "_normalized"
        if self.use_trunk_output:
            base_name += "_trunk"
        base_name += "_"+self.reference_set
        return base_name

    def accuracies_keyname(self, measure_name, label_level):
        measure_name += "_level%d" % label_level
        return measure_name

    def all_splits_combined(self, embeddings_and_labels):
        eee, lll = list(zip(*list(embeddings_and_labels.values())))
        curr_embeddings = np.concatenate(eee, axis=0)
        curr_labels = np.concatenate(lll, axis=0)
        return curr_embeddings, curr_labels

    def set_reference_and_query(self, embeddings_and_labels, curr_split):
        query_embeddings, query_labels = embeddings_and_labels[curr_split]
        if self.reference_set == "compared_to_self":
            reference_embeddings, reference_labels = query_embeddings, query_labels
        elif self.reference_set == "compared_to_sets_combined":
            reference_embeddings, reference_labels = self.all_splits_combined(embeddings_and_labels)
        elif self.reference_set == "compared_to_training_set":
            reference_embeddings, reference_labels = embeddings_and_labels["train"]
        else:
            raise BaseException 
        return query_embeddings, query_labels, reference_embeddings, reference_labels

    def embeddings_come_from_same_source(self, embeddings_and_labels):
        return self.reference_set in ["compared_to_self", "compared_to_sets_combined"]

    def record_group_name(self, split_name):
        return "%s_%s"%(self.base_record_group_name, split_name.upper())

    def do_knn_and_accuracies(self, embeddings_and_labels, accuracies, epoch, split_keys):
        raise NotImplementedError

    def test(self, dataset_dict, epoch, trunk_model, embedder_model, splits_to_eval=None, post_processor=None, collate_fn=None):
        logging.info("Evaluating epoch %d" % epoch)
        trunk_model = trunk_model.eval()
        embedder_model = embedder_model.eval()
        split_keys = list(dataset_dict.keys())
        embeddings_and_labels = {k: None for k in split_keys}
        for split_name, dataset in dataset_dict.items():
            logging.info('Getting embeddings for the %s split'%split_name)
            embeddings_and_labels[split_name] = self.get_all_embeddings(dataset, trunk_model, embedder_model, post_processor, collate_fn)
        splits_to_eval = split_keys if splits_to_eval is None else splits_to_eval
        for split_name in splits_to_eval:
            logging.info('Computing accuracy for the %s split'%split_name)
            accuracies = {"epoch":epoch}
            self.do_knn_and_accuracies(accuracies, embeddings_and_labels, epoch, split_name)
            if self.record_keeper is not None:
                self.record_keeper.update_records(accuracies, epoch, input_group_name_for_non_objects=self.record_group_name(split_name))
                best_epoch, best_accuracy = self.get_best_epoch_and_accuracy(split_name)
                accuracies = {"best_epoch":best_epoch, "best_accuracy":best_accuracy}
                self.record_keeper.update_records(accuracies, epoch, input_group_name_for_non_objects=self.record_group_name(split_name))
            else:
                logging.info(accuracies)


#! /usr/bin/env python3

from .base_miner import BasePostGradientMiner
from ..utils import loss_and_miner_utils as lmu
import torch


class MultiSimilarityMiner(BasePostGradientMiner):
    def __init__(self, epsilon, **kwargs):
        super().__init__(**kwargs)
        self.epsilon = epsilon

    def mine(self, embeddings, labels):
        self.n = embeddings.size(0)
        self.index_list = torch.arange(self.n).to(embeddings.device)
        self.sim_mat = lmu.sim_mat(embeddings)
        return self.compute_indices(labels)

    def compute_indices(self, labels):
        empty_tensor = torch.tensor([]).long().to(labels.device)
        a1_idx, p_idx, a2_idx, n_idx = [empty_tensor], [empty_tensor], [empty_tensor], [empty_tensor]
        for i in range(self.n):
            pos_indices = (
                ((labels == labels[i]) & (self.index_list != i)).nonzero().flatten()
            )
            neg_indices = (labels != labels[i]).nonzero().flatten()

            if pos_indices.size(0) == 0 or neg_indices.size(0) == 0:
                continue

            pos_sorted, pos_sorted_idx = torch.sort(self.sim_mat[i][pos_indices])
            neg_sorted, neg_sorted_idx = torch.sort(self.sim_mat[i][neg_indices])
            neg_sorted_filtered_idx = (
                (neg_sorted + self.epsilon > pos_sorted[0]).nonzero().flatten()
            )
            pos_sorted_filtered_idx = (
                (pos_sorted - self.epsilon < neg_sorted[-1]).nonzero().flatten()
            )

            pos_indices = pos_indices[pos_sorted_idx][pos_sorted_filtered_idx]
            neg_indices = neg_indices[neg_sorted_idx][neg_sorted_filtered_idx]

            if len(pos_indices) > 0:
                a1_idx.append(torch.ones_like(pos_indices) * i)
                p_idx.append(pos_indices)
            if len(neg_indices) > 0:
                a2_idx.append(torch.ones_like(neg_indices) * i)
                n_idx.append(neg_indices)

        return [torch.cat(idx) for idx in [a1_idx, p_idx, a2_idx, n_idx]]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pytorch_metric_learning",
    version="0.9.20",
    author="Kevin Musgrave",
    author_email="tkm45@cornell.edu",
    description="A flexible and extensible metric learning library, written in PyTorch.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/KevinMusgrave/pytorch_metric_learning",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
          'torch',
          'torchvision',
          'numpy',
          'tqdm'
    ],
)


#! /usr/bin/env python3


from .train_with_classifier import TrainWithClassifier


class CascadedEmbeddings(TrainWithClassifier):
    def __init__(self, embedding_sizes, logit_sizes=None, **kwargs):
        super().__init__(**kwargs)
        self.embedding_sizes = embedding_sizes
        self.logit_sizes = logit_sizes

    def calculate_loss(self, curr_batch):
        data, labels = curr_batch
        embeddings, labels = self.compute_embeddings(data, labels)
        s = 0
        for i, curr_size in enumerate(self.embedding_sizes):
            e = embeddings[:, s : s + curr_size]
            indices_tuple = self.maybe_mine_embeddings(e, labels)
            self.losses["metric_loss"] += self.maybe_get_metric_loss(e, labels, indices_tuple)
            self.update_record_keeper_in_loop(self.loss_funcs, ["metric_loss"], i)
            self.update_record_keeper_in_loop(self.mining_funcs, ["post_gradient_miner"], i)
            s += curr_size

        logits = self.maybe_get_logits(embeddings)
        if logits is not None:
            s = 0
            for i, curr_size in enumerate(self.logit_sizes):
                L = logits[:, s : s + curr_size]
                self.losses["classifier_loss"] += self.maybe_get_classifier_loss(L, labels)
                self.update_record_keeper_in_loop(self.loss_funcs, ["classifier_loss"], i)
                s += curr_size

    def update_record_keeper_in_loop(self, input_dict, keys, loop_iter):
        if self.record_keeper is not None:
            subset_dict = {}
            for k in keys:
                if k in input_dict:
                    subset_dict['%s_%d'%(k, loop_iter)] = input_dict[k]
            self.record_keeper.update_records(subset_dict, self.get_global_iteration())

    def record_these(self):
        # remove loss_funcs and mining_funcs since these are recorded in the loops above
        data_to_record = super().record_these()
        return data_to_record[:2] + data_to_record[4:]



from .angular_loss import AngularLoss, AngularNPairsLoss
from .contrastive_loss import ContrastiveLoss
from .generic_pair_loss import GenericPairLoss
from .lifted_structure_loss import GeneralizedLiftedStructureLoss
from .margin_loss import MarginLoss
from .multi_similarity_loss import MultiSimilarityLoss
from .nca_loss import NCALoss
from .n_pairs_loss import NPairsLoss
from .proxy_losses import ProxyNCALoss
from .triplet_margin_loss import TripletMarginLoss


#! /usr/bin/env python3


import torch
from ..utils import loss_and_miner_utils as lmu
from .base_metric_loss_function import BaseMetricLossFunction


class GenericPairLoss(BaseMetricLossFunction):
    """
    The function pair_based_loss has to be implemented by the child class.
    By default, this class extracts every positive and negative pair within a
    batch (based on labels) and passes the pairs to the loss function.
    The pairs can be passed to the loss function all at once (self.loss_once)
    or pairs can be passed iteratively (self.loss_loop) by going through each
    sample in a batch, and selecting just the positive and negative pairs
    containing that sample.
    Args:
        use_similarity: set to True if the loss function uses pairwise similarity
                        (dot product of each embedding pair). Otherwise,
                        euclidean distance will be used
        iterate_through_loss: set to True to use self.loss_loop and False otherwise
        squared_distances: if True, then the euclidean distance will be squared.
    """

    def __init__(
        self, use_similarity, iterate_through_loss, squared_distances=False, **kwargs
    ):
        self.use_similarity = use_similarity
        self.squared_distances = squared_distances
        self.loss_method = self.loss_loop if iterate_through_loss else self.loss_once
        super().__init__(**kwargs)

    def compute_loss(self, embeddings, labels, indices_tuple):
        mat = (
            lmu.sim_mat(embeddings)
            if self.use_similarity
            else lmu.dist_mat(embeddings, squared=self.squared_distances)
        )
        indices_tuple = lmu.convert_to_pairs(indices_tuple, labels)
        return self.loss_method(mat, labels, indices_tuple)

    def pair_based_loss(
        self, pos_pairs, neg_pairs, pos_pair_anchor_labels, neg_pair_anchor_labels
    ):
        raise NotImplementedError

    def loss_loop(self, mat, labels, indices_tuple):
        loss = torch.tensor(0.0).to(mat.device)
        n = 0
        (a1_indices, p_indices, a2_indices, n_indices) = indices_tuple
        for i in range(mat.size(0)):
            pos_pair, neg_pair = [], []
            if len(a1_indices) > 0:
                p_idx = a1_indices == i
                pos_pair = mat[a1_indices[p_idx], p_indices[p_idx]]
            if len(a2_indices) > 0:
                n_idx = a2_indices == i
                neg_pair = mat[a2_indices[n_idx], n_indices[n_idx]]
            loss += self.pair_based_loss(
                pos_pair, neg_pair, labels[a1_indices[p_idx]], labels[a2_indices[n_idx]]
            )
            n += 1
        return loss / (n if n > 0 else 1)

    def loss_once(self, mat, labels, indices_tuple):
        (a1_indices, p_indices, a2_indices, n_indices) = indices_tuple
        pos_pair, neg_pair = [], []
        if len(a1_indices) > 0:
            pos_pair = mat[a1_indices, p_indices]
        if len(a2_indices) > 0:
            neg_pair = mat[a2_indices, n_indices]
        return self.pair_based_loss(
            pos_pair, neg_pair, labels[a1_indices], labels[a2_indices]
        )
